@@ -1,20 +1,41 @@
 from __future__ import annotations
 
+import logging
 from pathlib import Path, PurePath
+import re
 from tempfile import TemporaryDirectory
-from typing import (
-    Any,
-    Dict,
-    Generic,
-    Iterator,
-    Literal,
-    Optional,
-    Tuple,
-    Type,
-    TypeVar,
-    Union,
-    get_args,
-)
+
+try:
+    from typing import (
+        Any,
+        Dict,
+        Generic,
+        Iterator,
+        Literal,
+        Optional,
+        Tuple,
+        Type,
+        TypeVar,
+        Union,
+        get_args,
+    )
+except ImportError:
+    from typing import (
+        Any,
+        Dict,
+        Generic,
+        Iterator,
+        Optional,
+        Tuple,
+        Type,
+        TypeVar,
+        Union,
+    )
+    from typing_extensions import (  # type: ignore
+        Literal,
+        get_args,
+    )
+
 from urllib.parse import unquote
 
 from openpyxl import load_workbook
@@ -33,9 +54,26 @@ import requests
 from .api import current_session
 from .utils import raise_for_status
 
+logger = logging.getLogger(__name__)
+
 FieldId = TypeVar('FieldId', bound=str)
 
 
+def shape_file_type(path: Path) -> Type[Segmentation] | Type[Mesh]:
+    """
+    Determine the type of the shape file.
+
+    As described in https://github.com/SCIInstitute/ShapeWorks/blob/3344bbfd42cb83eea50d01c06218e3034b5c67aa/Libs/Mesh/Mesh.h#L212,
+    Meshes are of file type "vtk", "vtp", "ply", "stl", "obj", while Segmentations are any other
+    file type.
+    """  # noqa: E501
+    file_type = path.suffix
+    if file_type in ['.vtk', '.vtp', '.ply', '.stl', '.obj']:
+        return Mesh
+    else:
+        return Segmentation
+
+
 class NonEmptyString(StrictStr):
     min_length = 1
 
@@ -106,9 +144,9 @@ class FileType(Generic[FieldId]):
             raise Exception('Unknown field id, this is likely a bug in the FileType class')
 
         with self.path.open('rb') as f:
-            self.field_value = session.s3ff.upload_file(f, self.path.name, self.field_id)[
-                'field_value'
-            ]
+            logger.info('Uploading file %s', self.path.name)
+            self.field_value = session.s3ff.upload_file(f, str(self.path.name), self.field_id)
+            logger.debug('Uploaded file %s', self.path.name)
 
         return self.field_value
 
@@ -130,6 +168,7 @@ class FileType(Generic[FieldId]):
         r = requests.get(self.url, stream=True)
         raise_for_status(r)
 
+        logger.info('Downloading %s', path)
         with path.open('wb') as f:
             for chunk in r.iter_content(chunk_size=8192):
                 f.write(chunk)
@@ -145,6 +184,9 @@ class FileType(Generic[FieldId]):
             return unquote(self.url.path.split('/')[-1])
         raise Exception('Invalid file object')
 
+    def __str__(self):
+        return self.name
+
 
 ModelType = TypeVar('ModelType', bound='ApiModel')
 
@@ -241,13 +283,14 @@ class ApiModel(BaseModel):
 
     def assert_local(self):
         if self.id is not None:
-            raise Exception('This entity already exists on the server.')
+            raise Exception('This entity already exists on the server (id: %r).' % self.id)
 
 
 class Dataset(ApiModel):
     _endpoint = 'datasets'
 
     name: NonEmptyString
+    file: Optional[FileType[Literal['core.Dataset.file']]] = None
     license: NonEmptyString
     description: NonEmptyString
     acknowledgement: NonEmptyString
@@ -255,6 +298,21 @@ class Dataset(ApiModel):
     contributors: str = ''
     publications: str = ''
 
+    def __repr_args__(self: BaseModel):
+        return [
+            (key, value)
+            for key, value in self.__dict__.items()
+            if key
+            not in {
+                'license',
+                'description',
+                'acknowledgement',
+                'keywords',
+                'contributors',
+                'publications',
+            }
+        ]
+
     @property
     def subjects(self) -> Iterator[Subject]:
         self.assert_remote()
@@ -274,6 +332,49 @@ class Dataset(ApiModel):
             for segmentation in subject.segmentations:
                 yield segmentation
 
+    @property
+    def meshes(self) -> Iterator[Mesh]:
+        for subject in self.subjects:
+            for mesh in subject.meshes:
+                yield mesh
+
+    @property
+    def images(self) -> Iterator[Image]:
+        for subject in self.subjects:
+            for mesh in subject.images:
+                yield mesh
+
+    def force_create(self, backup=False):
+        """
+        Forcibly create the Dataset, even if it already exists.
+
+        If backup=False (the default), then the existing Dataset is deleted.
+        If backup=True, then the new Dataset will append an appropriate `-v*` version string to
+        its name before creation.
+        """
+        old_dataset = Dataset.from_name(self.name)
+        while old_dataset is not None:
+            if not backup:
+                # Delete the old dataset to resolve the collision
+                old_dataset.delete()
+            else:
+                # Use a version suffix to resolve the collision
+                version_regex = re.compile('(.*)-v([0-9]+)')
+                match = version_regex.match(self.name)
+                if match:
+                    # The old name had a suffix, so increment it
+                    name, old_version = match.groups()
+                    logger.info('Trying a new name: %s, %s', name, old_version)
+                    new_version = int(old_version) + 1
+                    self.name = f'{name}-v{new_version}'  # type: ignore
+                else:
+                    # The old name had no suffix, so append "-v1"
+                    self.name = f'{self.name}-v1'  # type: ignore
+            # We have a new name now, but that new name might also conflict.
+            # Keep looping until there is no conflict.
+            old_dataset = Dataset.from_name(self.name)
+        return self.create()
+
     def add_project(self, file: Path, keywords: str = '', description: str = '') -> Project:
         project = Project(
             file=file,
@@ -281,7 +382,7 @@ class Dataset(ApiModel):
             description=description,
             dataset=self,
         ).create()
-        return project.load_project_spreadsheet()
+        return project._load_project_spreadsheet()
 
     @classmethod
     def from_name(cls, name: str) -> Optional[Dataset]:
@@ -291,45 +392,94 @@ class Dataset(ApiModel):
         except StopIteration:
             return None
 
-    def load_data_spreadsheet(self, file: Union[Path, str]) -> Dataset:
-        file = Path(file)
+    def create(self) -> Dataset:
+        result = super().create()
+        if self.file:
+            self._load_data_spreadsheet()
+        # Load the new dataset so we get an appropriate file field
+        assert result.id
+        return Dataset.from_id(result.id)
+
+    def _load_data_spreadsheet(self):  # noqa: C901
+        if not self.file or not self.file.path:
+            return
+        file = self.file.path
+
         xls = load_workbook(str(file), read_only=True)
         if 'data' not in xls:
             raise Exception('`data` sheet not found')
 
         data = xls['data'].values
 
-        if next(data)[0] != 'shape_file':
-            raise Exception('Unknown spreadsheet format')
+        headers = next(data)
+        for header in headers:
+            if not (header.startswith('shape_') or header.startswith('image_')):
+                raise Exception(
+                    f'Unknown spreadsheet format in {file} - expected "shape_" or "image_" prefix, found {header}'  # noqa: E501
+                )
+        # split each header into ('shape', anatomy_type) and ('image', modality) tuples
+        column_info = [header.split('_', 1) for header in headers]
+        # at least one shape column must be present
+        if all(info[0] != 'shape' for info in column_info):
+            raise Exception('No "shape_" column specified')
 
         root = file.parent
         subjects: Dict[str, Subject] = {}
 
+        found = False
         for row in data:
-            shape_file = root / row[0]
-            if not shape_file.exists():
-                raise Exception(f'Could not find shape file at "{shape_file}"')
+            subject = None
+            for i, cell in enumerate(row):
+                if not cell:
+                    continue
+                file_type, domain = column_info[i]
+                file_path: Path = root / cell
 
-            subject_name = shape_file.stem
-            if subject_name not in subjects:
-                subjects[subject_name] = self.add_subject(subject_name)
+                # Use the file name in the first cell as the subject name
+                if subject is None:
+                    subject_name = file_path.stem
+                    if subject_name not in subjects:
+                        subjects[subject_name] = self.add_subject(subject_name)
+                    subject = subjects[subject_name]
 
-            subject = subjects[subject_name]
-
-            # TODO: where to find the anatomy type?
-            subject.add_segmentation(file=shape_file, anatomy_type='unknown')
-
-        return self
+                if file_type == 'shape':
+                    shape_file = file_path
+                    if not shape_file.exists():
+                        raise Exception(f'Could not find shape file at "{shape_file}"')
+                    data_type = shape_file_type(shape_file)
+                    if data_type == Segmentation:
+                        subject.add_segmentation(file=shape_file, anatomy_type=domain)
+                    elif data_type == Mesh:
+                        subject.add_mesh(file=shape_file, anatomy_type=domain)
+                    found = True
+                elif file_type == 'image':
+                    image_file = file_path
+                    if not image_file.exists():
+                        raise Exception(f'Could not find image file at "{image_file}"')
+                    subject.add_image(file=image_file, modality=domain)
+                    found = True
+        if not found:
+            raise Exception('Did not find any shape or image files in data sheet')
 
     def download(self, path: Union[Path, str]):
         self.assert_remote()
         path = Path(path)
-        for segmentation in self.segmentations:
-            # TODO: add a dataset spreadsheet to the data model and get the path from it
-            segmentation.file.download(path / 'segmentations')
-
+        dataset_file = None
+        if self.file:
+            dataset_file = self.file.download(path)
         for project in self.projects:
             project.download(path)
+        if not dataset_file:
+            return
+        xls = load_workbook(str(dataset_file), read_only=True)
+        for subject in self.subjects:
+            row = next(entry for entry in xls['data'].values if subject.name == Path(entry[0]).stem)
+            for segmentation in subject.segmentations:
+                segmentation.file.download(path / Path(row[0]).parent)
+            for mesh in subject.meshes:
+                mesh.file.download(path / Path(row[0]).parent)
+            for image in subject.images:
+                image.file.download(path / Path(row[1]).parent)
 
 
 class Subject(ApiModel):
@@ -343,9 +493,25 @@ class Subject(ApiModel):
         self.assert_remote()
         return Segmentation.list(subject=self)
 
+    @property
+    def meshes(self) -> Iterator[Mesh]:
+        self.assert_remote()
+        return Mesh.list(subject=self)
+
+    @property
+    def images(self) -> Iterator[Image]:
+        self.assert_remote()
+        return Image.list(subject=self)
+
     def add_segmentation(self, file: Path, anatomy_type: str) -> Segmentation:
         return Segmentation(file=file, anatomy_type=anatomy_type, subject=self).create()
 
+    def add_mesh(self, file: Path, anatomy_type: str) -> Mesh:
+        return Mesh(file=file, anatomy_type=anatomy_type, subject=self).create()
+
+    def add_image(self, file: Path, modality: str) -> Image:
+        return Image(file=file, modality=modality, subject=self).create()
+
 
 class Segmentation(ApiModel):
     _endpoint = 'segmentations'
@@ -355,6 +521,22 @@ class Segmentation(ApiModel):
     subject: Subject
 
 
+class Mesh(ApiModel):
+    _endpoint = 'meshes'
+
+    file: FileType[Literal['core.Mesh.file']]
+    anatomy_type: NonEmptyString
+    subject: Subject
+
+
+class Image(ApiModel):
+    _endpoint = 'images'
+
+    file: FileType[Literal['core.Image.file']]
+    modality: str
+    subject: Subject
+
+
 class Project(ApiModel):
     _endpoint = 'projects'
 
@@ -368,6 +550,11 @@ class Project(ApiModel):
         self.assert_remote()
         return GroomedSegmentation.list(project=self)
 
+    @property
+    def groomed_meshes(self) -> Iterator[GroomedMesh]:
+        self.assert_remote()
+        return GroomedMesh.list(project=self)
+
     def add_groomed_segmentation(
         self,
         file: Path,
@@ -383,6 +570,21 @@ class Project(ApiModel):
             pre_alignment=pre_alignment,
         ).create()
 
+    def add_groomed_mesh(
+        self,
+        file: Path,
+        mesh: Mesh,
+        pre_cropping: Optional[Path] = None,
+        pre_alignment: Optional[Path] = None,
+    ) -> GroomedMesh:
+        return GroomedMesh(
+            file=file,
+            mesh=mesh,
+            project=self,
+            pre_cropping=pre_cropping,
+            pre_alignment=pre_alignment,
+        ).create()
+
     @property
     def shape_models(self) -> Iterator[OptimizedShapeModel]:
         self.assert_remote()
@@ -394,7 +596,7 @@ class Project(ApiModel):
             parameters=parameters,
         ).create()
 
-    def load_project_spreadsheet(self) -> Project:
+    def _load_project_spreadsheet(self) -> Project:
         file = self.file.path
         assert file  # should be guaranteed by assert_local
 
@@ -405,30 +607,45 @@ class Project(ApiModel):
             PurePath(segmentation.file.name).stem: segmentation
             for segmentation in self.dataset.segmentations
         }
+        meshes = {PurePath(mesh.file.name).stem: mesh for mesh in self.dataset.meshes}
 
-        if 'optimize' not in xls or 'data' not in xls:
-            raise Exception('`data` sheet not found')
+        # if 'optimize' not in xls or 'data' not in xls:
+        #     raise Exception('`data` sheet not found')
+
+        # shape_model = self._parse_optimize_sheet(xls['optimize'].values)
+        # self._parse_data_sheet(segmentations, shape_model, xls['data'].values, root)
+
+        for sheet_name in ['optimize', 'data']:
+            if sheet_name not in xls:
+                raise Exception(f'`{sheet_name}` sheet not found')
 
         shape_model = self._parse_optimize_sheet(xls['optimize'].values)
-        self._parse_data_sheet(segmentations, shape_model, xls['data'].values, root)
+        self._parse_data_sheet(segmentations, meshes, shape_model, xls['data'].values, root)
 
         return self
 
     def _iter_data_sheet(
         self, sheet: Any, root: Path
     ) -> Iterator[Tuple[Path, Path, str, Path, Path]]:
-        headers = next(sheet)
-        if headers != (
+        expected = (
             'shape_file',
             'groomed_file',
             'alignment_file',
             'local_particles_file',
             'world_particles_file',
-        ):
-            raise Exception('Unknown spreadsheet format')
+        )
+        headers = next(sheet)
+        if headers[: len(expected)] != expected:
+            raise Exception(
+                'Unknown spreadsheet format - expected headers to be %r, found %r'
+                % (expected, headers[: len(expected)])
+            )
 
         for row in sheet:
             shape_file, groomed_file, alignment_file, local, world = row
+            if not shape_file:
+                # It's possible to get XLSX files with empty rows where every column is None
+                continue
 
             shape_file = root / shape_file
             groomed_file = root / groomed_file
@@ -438,9 +655,13 @@ class Project(ApiModel):
             yield shape_file, groomed_file, alignment_file, local, world
 
     def _parse_optimize_sheet(self, sheet: Any) -> OptimizedShapeModel:
+        expected = ('key', 'value')
         headers = next(sheet)
-        if headers != ('key', 'value'):
-            raise Exception('Unknown spreadsheet format')
+        if headers[: len(expected)] != expected:
+            raise Exception(
+                'Unknown spreadsheet format - expected headers to be %r, found %r'
+                % (expected, headers[: len(expected)])
+            )
 
         params: Dict[str, Union[str, float]] = {}
         for row in sheet:
@@ -456,6 +677,7 @@ class Project(ApiModel):
     def _parse_data_sheet(
         self,
         segmentations: Dict[str, Segmentation],
+        meshes: Dict[str, Mesh],
         shape_model: OptimizedShapeModel,
         sheet: Any,
         root: Path,
@@ -463,14 +685,26 @@ class Project(ApiModel):
         for shape_file, groomed_file, alignment_file, local, world in self._iter_data_sheet(
             sheet, root
         ):
-            segmentation = segmentations.get(shape_file.stem)
-            if not segmentation:
-                raise Exception(f'Could not find segmentation for "{shape_file}"')
+            groomed_segmentation = None
+            groomed_mesh = None
+            data_type = shape_file_type(shape_file)
+            if data_type == Segmentation:
+                segmentation = segmentations.get(shape_file.stem)
+                if not segmentation:
+                    raise Exception(f'Could not find segmentation for "{shape_file}"')
+                groomed_segmentation = self.add_groomed_segmentation(
+                    file=groomed_file,
+                    segmentation=segmentation,
+                )
+            elif data_type == Mesh:
+                mesh = meshes.get(shape_file.stem)
+                if not mesh:
+                    raise Exception(f'Could not find mesh for "{shape_file}"')
+                groomed_mesh = self.add_groomed_mesh(
+                    file=groomed_file,
+                    mesh=mesh,
+                )
 
-            groomed_segmentation = self.add_groomed_segmentation(
-                file=groomed_file,
-                segmentation=segmentation,
-            )
             with TemporaryDirectory() as dir:
                 transform = Path(dir) / f'{shape_file.stem}.transform'
                 with transform.open('w') as f:
@@ -481,6 +715,7 @@ class Project(ApiModel):
                     local=local,
                     transform=transform,
                     groomed_segmentation=groomed_segmentation,
+                    groomed_mesh=groomed_mesh,
                 )
 
     def download(self, path: Union[Path, str]):
@@ -516,6 +751,17 @@ class GroomedSegmentation(ApiModel):
     project: Project
 
 
+class GroomedMesh(ApiModel):
+    _endpoint = 'groomed-meshes'
+
+    file: FileType[Literal['core.GroomedMesh.file']]
+    pre_cropping: Optional[FileType[Literal['core.GroomedMesh.pre_cropping']]] = None
+    pre_alignment: Optional[FileType[Literal['core.GroomedMesh.pre_alignment']]] = None
+
+    mesh: Mesh
+    project: Project
+
+
 class OptimizedShapeModel(ApiModel):
     _endpoint = 'optimized-shape-models'
 
@@ -531,7 +777,8 @@ class OptimizedShapeModel(ApiModel):
         world: Path,
         local: Path,
         transform: Path,
-        groomed_segmentation: GroomedSegmentation,
+        groomed_segmentation: Optional[GroomedSegmentation],
+        groomed_mesh: Optional[GroomedMesh],
     ) -> OptimizedParticles:
         return OptimizedParticles(
             world=world,
@@ -539,6 +786,7 @@ class OptimizedShapeModel(ApiModel):
             transform=transform,
             shape_model=self,
             groomed_segmentation=groomed_segmentation,
+            groomed_mesh=groomed_mesh,
         ).create()
 
 
@@ -549,7 +797,8 @@ class OptimizedParticles(ApiModel):
     local: FileType[Literal['core.OptimizedParticles.local']]
     transform: FileType[Literal['core.OptimizedParticles.transform']]
     shape_model: OptimizedShapeModel
-    groomed_segmentation: GroomedSegmentation
+    groomed_segmentation: Optional[GroomedSegmentation]
+    groomed_mesh: Optional[GroomedMesh]
 
 
 class OptimizedPCAModel(ApiModel):
